#include <stdio.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <string.h>
#include <time.h>
#include <curses.h>
#include <unistd.h>
#include <fcntl.h>
#include <getopt.h>
#include <errno.h>
#include <ctype.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <unistd.h>
#include <asm/types.h>
#include <linux/netlink.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <syslog.h>

// Control files - The existence of these files controls some behavior
// of the daemon.  This might get moved to using signals in the future
// Disable failure detection - this is vital if we're going to be sniffing!
const char disable_file[] = "/var/run/wlanwatcher-disable";

// Command template
const char mib_template[] = "%s wlan0 dot11req_mibget mibattribute=%s 2>&1";
// MIB's to get the connection quality and transfer rate
const char mib_quality[] = "p2CommsQuality";
// Promisc mode enable
const char promisc_mode[] = "%s wlan0 lnxreq_wlansniff channel=6 enable=%s";
// Path to the control application
const char control_path[] = "/sbin/wlanctl-ng";

// Template for starting it with the new options
const char opt_template[] = "/etc/pcmcia/wlan-switcher start wlan0 %s";
// Template for configuring the correct IP
const char ip_template[] = "/etc/rc.d/rc.inet.wlan0 %s";
// Option directory
const char opt_file[] = "/etc/pcmcia/wlan-ng.%s";

// Global error condition
int error = 0;

// Argument toggles
int toggle_debug = 0;
int toggle_linkabort = 0;
int toggle_syslog = 1;
int toggle_foreground = 0;
int toggle_verbose = 0;

// Argument values
int arg_linkdead = 30;

// General globals
int promisc_fd = 0;

// Data back from the mib get
typedef struct {
    int cq;
    int asl;
    int anl;
} PrismData;

// prismdump defines and structs
#define DEVNAME_LEN 16
typedef struct
{
  __u32 did __attribute__ ((packed));
  __u16 status __attribute__ ((packed));
  __u16 len __attribute__ ((packed));
  __u32 data __attribute__ ((packed));
}
p80211item_t;

typedef struct 
{
    struct timeval ts;
    __u32 caplen;
    __u32 len;
    // int pkt_encap;
} wtap_hdr_t;

typedef struct
{
    __u32 msgcode __attribute__ ((packed));
    __u32 msglen __attribute__ ((packed));
    __u8 devname[DEVNAME_LEN] __attribute__ ((packed));
    p80211item_t hosttime __attribute__ ((packed));
    p80211item_t mactime __attribute__ ((packed));
    p80211item_t channel __attribute__ ((packed));
    p80211item_t rssi __attribute__ ((packed));
    p80211item_t sq __attribute__ ((packed));
    p80211item_t signal __attribute__ ((packed));
    p80211item_t noise __attribute__ ((packed));
    p80211item_t rate __attribute__ ((packed));
    p80211item_t istx __attribute__ ((packed));
    p80211item_t frmlen __attribute__ ((packed));
} sniff_packet_t;

typedef struct
{
    __u16 frame_control __attribute__ ((packed));
    __u16 duration_id __attribute__ ((packed));
    __u8 mac1[6] __attribute__ ((packed));
    __u8 mac2[6] __attribute__ ((packed));
    __u8 mac3[6] __attribute__ ((packed));
    __u16 sequence __attribute__ ((packed));
    __u8 mac4[6] __attribute__ ((packed));
}
wlan_hdr_t;

typedef struct
{
    __u8 version:2;
    __u8 type:2;
    __u8 subtype:4;
    __u8 toDS:1;
    __u8 fromDS:1;
    __u8 morefrag:1;
    __u8 retry:1;
    __u8 pwr:1;
    __u8 moredata:1;
    __u8 wep:1;
    __u8 rsvd:1;
} frame_control_t;

typedef struct
{
    wtap_hdr_t wtap;
    wlan_hdr_t wlan;
} raw_packet_t;


// PacketSource globals
int pread, pwrite;

// Airsnort data, with PacketSource ported to C from C++
#define MCAST_GRP_SNIFF 2
#define ERR_IO -1
#define ERR_TIMEOUT -2 // timeout or cancel


// Open the netlink socket
int OpenPacketSource() {
    int fds[2], r;
    int fd;

    struct sockaddr_nl addr;
    r = pipe(fds);

    if (r < 0)
        return(r);

    pread = fds[0]; pwrite = fds[1];
    fd = socket(PF_NETLINK, SOCK_RAW, MCAST_GRP_SNIFF);

    if (fd < 0)
        return(r);

    addr.nl_family = PF_NETLINK;
    addr.nl_pid = getpid();
    addr.nl_groups = MCAST_GRP_SNIFF;

    if (bind(fd, (struct sockaddr *) &addr, sizeof(struct sockaddr_nl)) < 0)
        return(-1);

    return(fd);
}

// Get a packet from the netlink socket
int GetPacket(int fd, char *buf, int maxlen, int timeout) {
    fd_set rs;
    int r;
    struct timeval tm;
    struct timeval *ptm;
    if (pread < 0 || fd < 0) return(ERR_IO);
    FD_ZERO(&rs);
    FD_SET(fd, &rs);
    FD_SET(pread, &rs);

    if (timeout >= 0) {
        tm.tv_sec = timeout / 1000;
        tm.tv_usec = (timeout % 1000) * 1000;
        ptm = &tm;
    } else {
        ptm = NULL;
    }

    r=select(pread > fd ? pread + 1 : fd + 1, &rs, NULL, NULL, ptm);
    if (r < 0)
        return(ERR_IO);
    if (r == 0) {
        if (timeout >= 0)
            return(ERR_TIMEOUT);
    }

    if (FD_ISSET(pread, &rs)) {
        char a;
        read(pread, &a, 1);
    }

    if (FD_ISSET(fd, &rs)) {
        r=recv(fd, buf, maxlen,0);
        if (r < 0)
            return(ERR_IO);
        return(r);
    }

    return(ERR_TIMEOUT);
}


// Get the signal info from the prism control program
PrismData fetch_signaldata(void) {
    char cmdline[256];
    FILE *mib;
    char junkdat[256];
    PrismData dat = {-1, -1, -1};
    int cq,sl,sn;
    int scanret;

    snprintf(cmdline, 256, mib_template, control_path, mib_quality);

    // If we bombed return the (initialized) value of all negative
    if ((mib = popen(cmdline, "r")) == NULL) {
        if (toggle_debug)
            printf("ERROR:  Could not open control application %s.\n",
                   control_path);
        if (toggle_linkabort)
            exit(-1);

        error = 100;
        return dat;
    }

    // Discard the first line
    fscanf(mib, "%s", junkdat);

    // Get the data
    if (fscanf(mib, "  mibattribute=p2CommsQuality=%d,%d,%d",
               &cq, &sl, &sn) != 3) {
        if (toggle_debug)
            printf("ERROR:  Did not get expected data from wlanctl (command line: '%s')\n", cmdline);

        error = 1;
        pclose(mib);
        return dat;
    }

    // Fill it in and return
    dat.cq = cq;
    dat.asl = sl;
    dat.anl = sn;

    pclose(mib);
    return dat;
}

// Throw the card into promisc mode with wlanctl-ng
void enter_promisc(void) {
    char cmdline[256];
    FILE *mib;

    snprintf(cmdline, 256, promisc_mode, control_path, "true");

    // If we bombed return the (initialized) value of all negative
    if ((mib = popen(cmdline, "r")) == NULL) {
        if (toggle_debug)
            printf("ERROR:  Could not open control application %s.\n",
                   control_path);
        if (toggle_linkabort)
            exit(-1);

        error = 100;
    }

    promisc_fd = OpenPacketSource();

}

void exit_promisc(void) {
    char cmdline[256];
    FILE *mib;

    snprintf(cmdline, 256, promisc_mode, control_path, "false");

    // If we bombed return the (initialized) value of all negative
    if ((mib = popen(cmdline, "r")) == NULL) {
        if (toggle_debug)
            printf("ERROR:  Could not open control application %s.\n",
                   control_path);
        if (toggle_linkabort)
            exit(-1);

        error = 100;
    }

    if (promisc_fd)
        close(promisc_fd);
}

int handle_ssid(char *ssid) {
    char cmdline[256];
    int ret;
    struct stat statbuf;

    // Look to see if this is a SSID we have an option for
    snprintf(cmdline, 256, opt_file, ssid);
    if (stat(cmdline, &statbuf))
        return -1;

    // Run the (modified) wlan-ng start routine
    snprintf(cmdline, 256, opt_template, ssid);
    ret = system(cmdline);
    if (ret == -1 || ret == 127)
        return ret;

    snprintf(cmdline, 256, ip_template, ssid);
    ret = system(cmdline);
    if (ret == -1 || ret == 127)
        return ret;

    return 0;
}

// Look to see if we have the control file to disable us
int scan_disable() {
    struct stat statbuf;

    if (stat(disable_file, &statbuf))
        return 0;

    return -1;
}


// Main stuff
int main(int argc, char *argv[]) {
    int r, pid;
    PrismData link_status;
    int promisc = 0;
    time_t last_signal = time(0);

    sniff_packet_t *sniff_info;
    int len,oi;
    unsigned char packet[4000];
    char ssid[32];

    // Bits needed to find the ssid
    wlan_hdr_t *p80211_hdr;
    frame_control_t *fc;
    __u8 *msgbuf, temp;


    // Parse args
    while (1) {
        r = getopt(argc, argv, "sfavdl");

        if (r < 0)
            break;

        switch(r) {
        case 's':
            // Silent (no syslogging)
            toggle_syslog = 0;
            break;
        case 'd':
            // Debug mode
            toggle_debug = 1;
            break;
        case 'f':
            // Foreground
            toggle_foreground = 1;
            break;
        case 'a':
            // Abort if link file unavailable
            toggle_linkabort = 1;
            break;
        case 'v':
            // Verbose
            toggle_verbose = 1;
            break;
        case 'l':
            // Set linkdead timeout
            if (argc < optind + 1) {
                printf("%s - Expected argument for -l, see --help\n",
                       argv[0]);
                exit(1);
            }

            if (!isdigit(argv[optind][0])) {
                printf("%s - Expected numerical argument for -l, see --help\n",
                       argv[0]);
                exit(1);
            }

            arg_linkdead = atoi(argv[optind]);

            break;
        default:
            // Usage
            fprintf(stderr, "Usage:  %s [-s] [-f] [-a] [-v] [-l #]\n"
                    "    -s     Silent (don't syslog when things happen)\n"
                    "    -f     Foreground (non-daemon mode)\n"
                    "    -a     Abort if the link file becomes unavailable\n"
                    "    -v     Verbose operation\n"
                    "    -d     Debug mode (command output)\n"
                    "    -l #   Linkdeath timeout before searching (default: 30)\n",
                    argv[0]);
            exit(1);
            break;
        }
    }

    if (toggle_syslog) {
        openlog(argv[0], LOG_PID, LOG_DAEMON);
        syslog(LOG_INFO, "Wlanwatch monitoring wlan0");
    }

    // Fork and let the foreground copy of us die
    if (!toggle_foreground) {
        if ((pid = fork()) < 0) {
            if (toggle_syslog)
                syslog(LOG_INFO, "Failed to fork to background.");
            perror("Couldn't fork to background");
        }

        if (pid > 0)
            exit(1);
    }


    while (1) {
        sleep(1);

        error = 0;

        // If we're not sniffing for SSID's
        link_status = fetch_signaldata();

        // If we're disabled...
        if (scan_disable() != 0) {
            if (toggle_verbose)
                printf("%s - Disabled due to control file.\n", argv[0]);

            // Drop out of promisc if we're in it
            if (promisc)
                exit_promisc();

            promisc = 0;

            // Go to sleep for a long time
            sleep(10);

            continue;

        }


        if (error) {
            // If we don't have a card, sleep extra
            if (toggle_verbose)
                printf("%s - No card found\n", argv[0]);

            promisc = 0;
            sleep(5);
            continue;
        }

        // Detect a link failure and go into promisc mode if we're not already
        // there
        if (link_status.cq == 0 && link_status.asl == 27 &&
            link_status.anl == 27) {
            // If we don't seem to have a link
            if (((time(0) - last_signal) > arg_linkdead) && !promisc) {
                if (toggle_verbose)
                    printf("%s - No signal for more than %d seconds.  Going promisc.\n",
                           argv[0], arg_linkdead);
                if (toggle_syslog)
                    syslog(LOG_INFO, "No signal for more than %d seconds.  Going promisc.",
                           arg_linkdead);
                enter_promisc();
                promisc = 1;
            }
        } else {
            // Otherwise we have at least a link-layer connection, so
            // tag the current time and sleep a little extra
            last_signal = time(0);
            sleep(5);
            promisc = 0;
        }

        // Sniff the packets
        if (promisc) {
            if (toggle_debug)
                printf("%s - Fetching promisc packet.\n", argv[0]);

            bzero(packet, 4000);

            // We're sniffing for ssid's -- handle the incoming packets
            len = GetPacket(promisc_fd, (char *) packet, 4000, 1000);

            if (len == ERR_TIMEOUT) {
                if (toggle_verbose)
                    printf("%s - Timed out fetching packet.\n", argv[0]);
            }

            if( errno == ENOBUFS ) continue ;


            if (toggle_debug)
                printf("%s - Extracting ssid.\n", argv[0]);

            bzero(ssid, 32);
            oi = sizeof (sniff_packet_t);
            p80211_hdr = (wlan_hdr_t *) &packet[oi];
            fc = (frame_control_t *) &p80211_hdr->frame_control;
            msgbuf = (__u8 *) p80211_hdr;

            if ((fc->type == 0) && ( fc->subtype == 8)) // just beacon packets
            {
                temp=(msgbuf[37] & 0xff)+1;
                snprintf(ssid, temp, "%s", &msgbuf[38]);

                // Handle it
                if (toggle_verbose)
                    printf("%s - Saw SSID '%s'\n", argv[0], ssid);

                if (!handle_ssid(ssid)) {
                    if (toggle_verbose)
                        printf("%s - Switched to SSID '%s'\n", argv[0], ssid);
                    if (toggle_syslog)
                        syslog(LOG_INFO, "Switched to SSID '%s'\n", ssid);

                    exit_promisc();
                    last_signal = time(0);
                    promisc = 0;
                }

            }

        }
    }

}